////////////////////////////////////////////////////
// file: RS_behavior.cpp                          //
//                                                //
// task: shows how to program a behavior          //
//       for the RoboSapien robot by using        //
//       the robot.cpp and camera.cpp class.      //
//       As an example we show how to program     //
//       the robot running towards an orange      //
//       pole.                                    //
//                                                //
// authors: Team NimbRo,                          //
//          Albert-Ludwigs-University of Freiburg //
//          Germany                               //
//          www.nimbro.net                        //
//          September 2004                        //
//                                                //
// compile: the source-code was programmed and    //
//          compiled with Microsoft's eMbedded    //
//          Visual C++ 3.0 IDE as a part of the   //
//          eMbedded Visual Tools 3.0 - 2002      //
//          Edition. Can be downloaded and used   //
//          for free from www.microsoft.com       //
////////////////////////////////////////////////////



///////////////////
// include files //
///////////////////

// for video capture
#include "camera.h"

// which camera types are available?
#include "camera_types.h"

// for moving the RoboSapien
#include "robot.h"

// Pocket PC stuff
#include <aygshell.h>



/////////////////
// definitions //
/////////////////

// dimensions of the picture to be captured (1280x1024 or 640x480 or 320x240 or 160x120)
#define PIC_WIDTH 160
#define PIC_HEIGHT 120

// for the FlyCAM-camera the display and the capture dimensions differ
int DISPLAY_WIDTH = PIC_WIDTH;
int DISPLAY_HEIGHT = PIC_HEIGHT;

// where shall we paint the picture?
#define DISPLAY_PICTURE_AT_X 5
#define DISPLAY_PICTURE_AT_Y 10

// IDs of text labels
#define IDC_STATUS  1000
#define IDC_STATUS2 1001
#define IDC_STATUS3 1002

// menu bar height
#define MENU_HEIGHT 26



//////////////////////
// global variables //
//////////////////////

// for video capture we need access to the camera
camera* p_myCamera;

// a pointer to a byte array storing the picture
PBYTE p_PicBuffer = NULL;

// for controlling the RoboSapien
robot* p_myRobot;

// window title
LPTSTR szTitle = L"Example Behavior";
LPTSTR szClass = L"RS_behavior";

// instance handle
HINSTANCE ghInst = NULL;

// window handle
HWND ghWnd = NULL;

// handle to the device dependent bitmap
HDC hdcBitmap = NULL;

// handle to the DIB-bitmap
HBITMAP hBitmap = NULL;

// handle to the display thread
HANDLE hThread = NULL;

// shall the display thread be stopped?
BOOL EndThread = FALSE;

// move the robot or only do video capturing?
BOOL MOVE_ROBOT = TRUE;



//////////////////////////////////////
// method: CreateBitmap             //
// task  : creates a bitmap that is //
//         build of the data that   //
//         p_PicBuffer points to.   //
//////////////////////////////////////
BOOL CreateBitmap()
{
	
	// get a handle to the display device context (DC)
	HDC hdc = GetDC(0);

	// create a bitmap info header
	BITMAPINFOHEADER bmih;
	bmih.biSize          = sizeof (BITMAPINFOHEADER) ;
	bmih.biWidth         = DISPLAY_WIDTH;
	bmih.biHeight        = DISPLAY_HEIGHT;
	bmih.biPlanes        = 1 ;
	bmih.biBitCount      = 24 ;
	bmih.biCompression   = BI_RGB ;
	bmih.biSizeImage     = 0 ;
	bmih.biXPelsPerMeter = 0 ;
	bmih.biYPelsPerMeter = 0 ;
	bmih.biClrUsed       = 0 ;
	bmih.biClrImportant  = 0 ;

	// Create a device independet Bitmap (DIB)
	hBitmap = CreateDIBSection (NULL, (BITMAPINFO *)  &bmih, 0, (VOID **)&p_PicBuffer, NULL, 0);

	// was CreateDIBSection successfull?
	if (!hBitmap)
	{
		// no, so release the device context
		ReleaseDC(0,hdc);
		return FALSE;
	}

	// create a memory device context
	hdcBitmap = CreateCompatibleDC(hdc);

	// select an object for the device context
	SelectObject(hdcBitmap,hBitmap);

	// no release the device context of the current window
	ReleaseDC(0,hdc);

	// everything is fine
	return TRUE;

} // end-method CreateBitmap



//////////////////////////////////////
// method: DeleteBitmap             //
// task  : release the bitmap and   //
//         device bitmap handle.    //
//////////////////////////////////////
void DeleteBitmap()
{
	if (hdcBitmap)
	{
		DeleteDC(hdcBitmap);
		hdcBitmap = NULL;
	}
	
	if (hBitmap)
	{
		DeleteObject(hBitmap);
		hBitmap = NULL;
	}

} // end-method DeleteBitmap



//////////////////////////////////////
// method: setPixel                 //
// task  : paints a pixel into the  //
//         picture at a certain     //
//         position.                //
//////////////////////////////////////
void setPixel(int x,int y,int R,int G,int B) {

	int byteArrayPos = 3 * (x+y*DISPLAY_WIDTH);
	p_PicBuffer[byteArrayPos]   = B;
	p_PicBuffer[byteArrayPos+1] = G;
	p_PicBuffer[byteArrayPos+2] = R;

} // end-method setPixel



//////////////////////////////////////
// method: ComputeCenterOfOrange    //
// task  : computes the average x   //
//         and y coordinate of all  //
//         pixels in a picture that //
//         are classified to be     //
//         orange.                  //
//////////////////////////////////////
void ComputeCenterOfOrange(int* _p_centerX, int* _p_centerY, int* _p_nrOrangePixels) {

	int byteArrayPos;
	int R,G,B;
	*_p_centerX=0;
	*_p_centerY=0;
	*_p_nrOrangePixels=0;
		
	// search for orange pixels in the picture
	for (int y=0; y<DISPLAY_HEIGHT; y++) {
		for (int x=0; x<DISPLAY_WIDTH; x++) {

			// get the RGB-values of the pixel at (x,y) in the picture
			byteArrayPos = 3 * (x+y*DISPLAY_WIDTH);
			B = p_PicBuffer[byteArrayPos];
			G = p_PicBuffer[byteArrayPos+1];
			R = p_PicBuffer[byteArrayPos+2];
					
			// can the RGB-values be considered as orange?
			if (R>G+20 && G>B+10) {

				// dye the pixel into a luscious orange
				p_PicBuffer[byteArrayPos]   = 0;
				p_PicBuffer[byteArrayPos+1] = 0;
				p_PicBuffer[byteArrayPos+2] = 255;

				// one pixel more classified as orange
				(*_p_nrOrangePixels)++;
				(*_p_centerX) += x;
				(*_p_centerY) += y;
											
			} // end-if
										
		} // for x
	} // for y

	// compute center of all orange pixels
	if ( *_p_nrOrangePixels>0 ) {
		*_p_centerX = (*_p_centerX) / (*_p_nrOrangePixels);
		*_p_centerY = (*_p_centerY) / (*_p_nrOrangePixels);


		// paint a cross-hairs at the point where we think the center of orange is
		for (int i=1; i<=11; i++) {
			setPixel( (*_p_centerX)-5+i, (*_p_centerY)     , 255,255,255 );
			setPixel( (*_p_centerX)    , (*_p_centerY)-5+i , 255,255,255 );
		} // end-for

	} else {

		// center of orange is not valid
		*_p_centerX = -1;
		*_p_centerY = -1;
	} // end-if

			
} // end-method ComputeCenterOfOrange



//////////////////////////////////////
// method: BehaviorThread           //
// task  : captures images, tries   //
//         to find the orange pole  //
//         and determines which     //
//         movement to perform to   //
//         bring the robot nearer   //
//         to the orange pole.      //
//////////////////////////////////////
DWORD WINAPI BehaviorThread(LPVOID lpParam)
{

	// prepare the video capturing. decomment the corresponding line depending on the type of your camera
	p_myCamera = new camera();

	#ifdef LIFEVIEW_FLYCAM
		DISPLAY_WIDTH = PIC_HEIGHT;
		DISPLAY_HEIGHT = PIC_WIDTH;
	#endif
	
	// init camera API and set capture dimensions
	p_myCamera->init(PIC_WIDTH,PIC_HEIGHT);

	// create a robot object
	p_myRobot = new robot();
	
	// handle to the device context
	HDC hdc;

	// for calculating the frame rate
	int cFrames = 0;

	// whats the time?
	DWORD StartTime = GetTickCount();

	// for displaying messages
	TCHAR szMsg[64];

	// create a bitmap for displaying pictures
	CreateBitmap();

	// test sequence for the upper arm
	/*
	p_myRobot->MoveUpperArm(Lowest,Left);
	p_myRobot->MoveUpperArm(Lower,Left);
	p_myRobot->MoveUpperArm(Middle,Left);
	p_myRobot->MoveUpperArm(Upper,Left);
	p_myRobot->MoveUpperArm(Lowest,Left);
	*/

	// test sequence for the lower arm
	/*
	p_myRobot->MoveLowerArm(Outer,Left);
	p_myRobot->MoveLowerArm(Inner,Left);
	p_myRobot->MoveLowerArm(Center,Left);
	*/

	// test sequence for the tilt move
	/*
	p_myRobot->TiltTrunk(Left);
	p_myRobot->TiltTrunk(Right);
	p_myRobot->TiltTrunk(Left);
	p_myRobot->TiltTrunk(Left);
	*/
	
	// as long as the behavior thread shall not terminate ...
	while (!EndThread)
	{
		// get the device context
		hdc = GetDC(ghWnd);

		// did we get it?
		if (hdc)
		{			

			// video capturing from the camera
			p_myCamera->getPicture(p_PicBuffer);

			// analyze the picture: where is the orange pole?
			int centerX=0;
			int centerY=0;
			int orangePixels=0;
			ComputeCenterOfOrange(&centerX, &centerY, &orangePixels);
			swprintf(szMsg,L"Pole x:%d, y:%d, pixel:%d",centerX,centerY,orangePixels);
			SetDlgItemText(ghWnd,IDC_STATUS,szMsg);

			if (MOVE_ROBOT) {

				/////////////////////////////////////////////////////////
				// compute a movement command for RoboSapien depending //
				// on the position of the orange pole in the picture   //
				/////////////////////////////////////////////////////////

				// the orange pole is to far to the right -> turn right
				if (centerX<(int) (PIC_WIDTH*0.30)) {
					swprintf(szMsg,L"Turning RIGHT");
					p_myRobot->Turn(15);
				}

				// the orange pole is to far to the left -> turn left
				if (centerX>(int) (PIC_WIDTH*0.70)) {
					swprintf(szMsg,L"Turning LEFT");
					p_myRobot->Turn(-15);

				}

				// the orange pole is approximately in the center of the picture
				// -> just walk forward
				if (centerX>=(int) (PIC_WIDTH*0.30) && centerX<=(int) (PIC_WIDTH*0.70)) {
					swprintf(szMsg,L"WALKING");
					p_myRobot->Move(15,STEPTYPE_BIGSTEPS);
				}
				SetDlgItemText(ghWnd,IDC_STATUS2,szMsg);

			} // end-if

			// show picture: copy bitmap from source device context (hdcBitmap) to device context (hdc)
			BitBlt( hdc,
				    DISPLAY_PICTURE_AT_X,
				    DISPLAY_PICTURE_AT_Y,
					DISPLAY_WIDTH,
					DISPLAY_HEIGHT,
					hdcBitmap,
					0,
					0,
					SRCCOPY);
						
			// release the device context
			ReleaseDC(ghWnd,hdc);
			
			// Work out framerate
			cFrames++;
			if (cFrames == 10)
			{
				swprintf(szMsg,L"Framerate: %.2f fps",(float)10*1000/(GetTickCount()-StartTime));
				SetDlgItemText(ghWnd,IDC_STATUS3,szMsg);
				StartTime = GetTickCount();
				cFrames = 0;
			} // end-if

			
		} // end-if (we got the device context)

	} // end-while (behavior thread shall not terminate)

		// close the camera
	p_myCamera->close();

	return 0;

} // end-method BehaviorThread


///////////////////////////////////////
// method: BehaviorStart             //
// task  : start the behavior thread //
///////////////////////////////////////
BOOL BehaviorStart(HWND hWnd)
{	
	DWORD ThreadID;

	// check whether the display thread is already running
	if (hThread) return FALSE;

	EndThread = FALSE;

	hThread = CreateThread(NULL,0,BehaviorThread,NULL,0,&ThreadID);

	return TRUE;

} // end-method BehaviorStart


///////////////////////////////////////
// method: BehaviorStop              //
// task  : stop the behavior thread  //
///////////////////////////////////////
BOOL BehaviorStop()
{
	if (!hThread)
		return FALSE;

	EndThread = TRUE;

	if (WaitForSingleObject(hThread,2000)==WAIT_TIMEOUT)
		MessageBeep(MB_ICONEXCLAMATION);

	CloseHandle(hThread);
	hThread = NULL;

	return TRUE;

} // end-method BehaviorStop


//////////////////////////////////////
// method: WndProc                  //
// task  : process windows events   //
//////////////////////////////////////
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	// which message did we receive?
	switch (message)
	{
		// build up the window
		case WM_CREATE:
			CreateWindow(L"STATIC",L"Status",WS_VISIBLE|WS_CHILD,10,230,200,20,hWnd,(HMENU)IDC_STATUS, ghInst,NULL);
			CreateWindow(L"STATIC",L"Status",WS_VISIBLE|WS_CHILD,10,252,200,20,hWnd,(HMENU)IDC_STATUS2,ghInst,NULL);
			CreateWindow(L"STATIC",L"Status",WS_VISIBLE|WS_CHILD,10,274,200,20,hWnd,(HMENU)IDC_STATUS3,ghInst,NULL);
			break;

		// re-activate window
		case WM_ACTIVATE:
			if (LOWORD(wParam) != WA_INACTIVE)
				SHFullScreen(hWnd,SHFS_HIDESIPBUTTON);
			break;

		// ok clicked
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
			case IDOK:
				DestroyWindow(hWnd);
				break;
			}
			break;

		// destroy window
		case WM_DESTROY:
			BehaviorStop();
			DeleteBitmap();
			PostQuitMessage(0);
			break;

		// pass window event
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);

	} // end-switch

	return 0;

} // end-method WndProc



////////////////////////////////////////
// method: WinMain                    //
// task  : build up the window and    //
//         start the behavior thread. //
////////////////////////////////////////
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPTSTR    lpCmdLine,
					int       nCmdShow)
{
	
	ghInst = hInstance;

	// look for a running instance of this application
	HWND hWnd;
	hWnd = FindWindow(szClass, szTitle);
	if (hWnd)
	{
		// there is already one! bring it to the foreground
		SetForegroundWindow(hWnd);
		return FALSE;
	} // end-if
		
	// register the class for subsequent use in the window
	WNDCLASS wc;
	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= (WNDPROC) WndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hInstance;
	wc.hIcon			= 0;
	wc.hCursor			= 0;
	wc.hbrBackground	= (HBRUSH) GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName		= 0;
	wc.lpszClassName	= szClass;
	RegisterClass(&wc);

	// where to create the window?
	int x = GetSystemMetrics(SM_CXSCREEN);
	int y = GetSystemMetrics(SM_CYSCREEN)-MENU_HEIGHT;

	// try to create a window with an extended style
	ghWnd = CreateWindowEx(WS_EX_CAPTIONOKBTN,szClass, szTitle, WS_VISIBLE,
		0, MENU_HEIGHT, x, y, NULL, NULL, hInstance, NULL);

	// did it worked?
	if (!ghWnd)	return 1;
	
	// show the window
	ShowWindow(ghWnd, nCmdShow);
	UpdateWindow(ghWnd);

	// start the behavior thread
	BehaviorStart(hWnd);

	// do message processing
	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	return msg.wParam;

} // end-method WinMain